/*
 *  Copyright(c)2002 Forward Computing and Control Pty. Ltd.
 *  ACN 003 669 994, NSW, Australia      All rights Reserved
 *  
 *   Written by Dr. M.P. Ford
 *
 *  ThreadReturn package Copyright (c) 2003 Forward Computing and Control Pty. Ltd.
 *  NSW. Australia,  www.forward.com.au
 *  All rights reserved.
 *  
 *  Redistribution of the source of this package, with or without
 *  modification, is NOT permitted.
 *  
 *  Redistribution of the compiled java .class files of this package, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions in compiled java .class form must reproduce the above
 *     copyright notice, this list of conditions and the following disclaimer
 *     in the documentation and/or other materials provided with the distribution.
 *  2. If the software is modified, the compiled java .class distribution must
 *     include, in the documentation and/or other materials provided with the 
 *     distribution, a statement that the software has been modified and
 *     identify the person who was responsable for the modifications.  
 *  3. The name of the author may not be used to endorse or promote products
 *     derived from this software without specific prior written permission.
 *  
 *  THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 *  IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 *  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */
package au.com.forward.threads;
import java.util.Date;

/**
 * This class sets a timeout time which can by used by mulitple join() statements to
 * specify the total time required for a number of threads to complete.
 * Typical use is:-
 *
<pre>
thread_1.start();  // start all the threads
thread_2.start();
thread_3.start();

TimeOut timeOut = new TimeOut(5000); // time out expires in 5 seconds from now

ThreadReturn.join(thread_1,timeOut.timeRemaining()); 
ThreadReturn.join(thread_2,timeOut.timeRemaining()); 
ThreadReturn.join(thread_3,timeOut.timeRemaining()); 
</pre>
 * 
 * This sets the timeout at 5 seconds from the time the <code>TimeOut</code> constructor
 * is executed.  This sets an absolute end time of now + 5 seconds.
 * <p>
 * The first <code>join()</code> statement receives timeout arguement of the absolute end time - now in
 * milliseconds. When this <code>join()</code> returns the following <code>join()</code> receives 
 * a timeout arguement of only the remaining time (in milliseconds) and so on.
 * This ensures that each of the three threads have
 * at most 5 seconds to complete their tasks.
 * <p>
 * If the absolute end time has been passed then <code>timeRemaining()</code> returns 1 millisecond as a minimum value
 * because 0 causes <code>join()</code> to wait forever and negative numbers are not valid arguements for
 * <code>join()</code>. This means that maximum timeout is actually 5 sec + 2 milliseconds for this example
 * or in general<br>
 * maximum timeout = constructor timeout + (no. of thread joins - 1) * 1mS
 *
 *
 * @author     matthew ford 
 * @created    7 April 2002 
 */
public class TimeOut extends Date {

	/**
	 *  Creates a new timeout with an absolute time of now + <code>milliseconds</code>
	 *
	 * @param  milliseconds  the timeout. 
	 */
	public TimeOut(long milliseconds) {
		super();
		long time = getTime();
		setTime(time+milliseconds);
	}


	/**
	 * Returns the number of milliseconds remaining from now.
	 *
	 * @return  the number of milliseconds remaining from now or 1 if none left
	 */
	public long timeRemaining() {
		long time = getTime() - (new Date()).getTime();
		if (time <= 0) {
			time = 1;
		} 
		return time;
	}


}